////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_DEFINEREFLECTION_H_INC_
#define REFLIB_DEFINEREFLECTION_H_INC_

#include <boost/type_traits.hpp>

#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/add_reference.hpp>
#include <boost/pending/ct_if.hpp>

#include "../recordtags.h"

#include "../initreflectedmembers.h"

#include "../priviligedaccess.h"

#include "../isruntimeiterable.h"

#include "../userconfig.h"


#ifdef _DEBUG
#include <typeinfo>
// may help to debug template code
#endif

#ifdef REFLIB_MEMVARS_WITH_TRAILING_
#define REFLIBMEMVARNAME(TagName) TagName##_
#else 
#define REFLIBMEMVARNAME(TagName) m_##TagName
#endif

//#define FORMAT_PREPROCESSED_OUTPUT

#ifdef FORMAT_PREPROCESSED_OUTPUT
#define REFLIB_LF ctrl_h_linefeed
#else
#define REFLIB_LF
#endif
namespace refl
{
    template<class ReflectedClass,int Pos> struct GetAttributeAt
    {
        typedef typename ReflectedClass::template GetNameTag<Pos>::type
            AttributeTag;

        typedef typename ReflectedClass::
            template GetValueType<Pos>::const_ref_type AttributeConstRef;
        
        typedef typename ReflectedClass::
            template GetMemberKind<Pos>::type AttributeKind;

        static AttributeConstRef getValue(const ReflectedClass& record)
        {
            return record.template getByTag<AttributeTag>();
        }
    };

    // - AttributeKind will be the kind of member variable:
    // -    either: InstanceVariableTag
    // -    or:     ClassVariableTag
    template<class ReflectedClass, class ValueType, class AttributeKind,
        class AttributeTag> class SetWithPreAndPost
    {
    public:
        typedef typename ::boost::add_const<ValueType>::type const_type;
        typedef typename ::boost::add_reference<const_type>::type 
            const_ref_type;  
        typedef void (ReflectedClass::*SetFunction)(const_ref_type,
                                                    AttributeTag);
        typedef typename userconfig::LockType<ReflectedClass,
                                    userconfig::ThreadingModel>::type
            Lock;
        SetWithPreAndPost(ReflectedClass& host, SetFunction setter, 
                        const_ref_type newVal):
                host_(host),setter_(setter),newVal_(newVal)
        { 
            userconfig::preSet(host_,AttributeTag(),newVal);
        }
        void operator()()const
        {
            {
                Lock lock(host_.getMutex());
                (host_.*setter_)(newVal_,AttributeTag());
            }   // don't lock postSet
            userconfig::postSet(host_,AttributeTag());  
                                   // not in the destructor because
                                   // we want to have the option to
        }                          // throw in postSet
    private:
        ReflectedClass& host_;
        SetFunction setter_;
        const_ref_type newVal_;
    };
    // To support multithreading you might want to add an appropriate Mutex
    // as a member variable to this template specialization.
    template<class ReflectedClass, class ValueType,
        class AttributeTag> class SetWithPreAndPost<ReflectedClass,ValueType,
                                        ClassVariableTag,AttributeTag>
    {
    public:
        typedef typename ::boost::add_const<ValueType>::type const_type;
        typedef typename ::boost::add_reference<const_type>::type 
            const_ref_type;
        typedef void (*SetFunction)(const_ref_type,AttributeTag);
        typedef typename userconfig::LockType<ReflectedClass,
                                            userconfig::ThreadingModel>::type
            Lock;

        SetWithPreAndPost(SetFunction setter,
                        const_ref_type newVal):
                setter_(setter),newVal_(newVal)
        {
            userconfig::preSet<ReflectedClass,AttributeTag,const_ref_type>(
                                                        AttributeTag(),newVal);
        }
        void operator()()const
        {
            {
                Lock lock(userconfig::MutexInstance<ReflectedClass,
                                            userconfig::ThreadingModel>::get());
                (*setter_)(newVal_,AttributeTag());
            }
            userconfig::postSet<ReflectedClass,AttributeTag>(AttributeTag());
            // not in the destructor because we want to keep the option to
        }   // throw in postSet
    private:
        SetFunction setter_;
        const_ref_type newVal_;
    };
    // T can be a class that is not copyable but is default constructible
    // (like boost::mutex is).
    // This template enables having a mutex inside a class that should be 
    // copyable 
    template<class T> struct UniqueResource
    {
        UniqueResource(){}
        UniqueResource(const UniqueResource& )
        {/* no state to copy - a default constructed T inside *this just does it.*/}
        UniqueResource& operator=(const UniqueResource&)
        {
            /* no state to assign - leave everything in place.*/
            return *this;
        }
        T resource_;
    };
}

// to get a diagnostic message, when an attempt is made to access a non existing
// reflected member.
struct ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END;

struct ERROR_ERROR_ERROR_no_appropriate_setter_defined_ERROR_END;
// use this form if you don't the library to define a default ctor.
#define BEGIN_REFLECTION_NO_CTOR(ClassName)                          REFLIB_LF \
public:                                                              REFLIB_LF \
    typedef userconfig::MutexType<ClassName,                         REFLIB_LF \
                                userconfig::ThreadingModel>::type    REFLIB_LF \
        MutexType;                                                   REFLIB_LF \
    refl::UniqueResource<MutexType> mutex_;                          REFLIB_LF \
    MutexType& getMutex()                                            REFLIB_LF \
    {   return mutex_.resource_;}                                    REFLIB_LF \
    friend struct refl::detail::PrivilegedAccessImpl;                REFLIB_LF \
    /* while we are at it we can do this as well - note that the   */REFLIB_LF \
    /* C++ Standard does not specify the contents of typeid(Class) */REFLIB_LF \
    /* name() and some implementations supply technical information*/REFLIB_LF \
    /* (namespaces, struct/class as well).                         */REFLIB_LF \
    static const std::string& getClassName()                         REFLIB_LF \
    {                                                                REFLIB_LF \
        static const std::string className(#ClassName);              REFLIB_LF \
        return className;                                            REFLIB_LF \
    }                                                                REFLIB_LF \
    void initReflectedMembers()                                      REFLIB_LF \
    {                                                                REFLIB_LF \
        refl::initReflectedMembers(*this);                           REFLIB_LF \
    }                                                                REFLIB_LF \
    /* A little dirty trick to be able to detect at compile time   */REFLIB_LF \
    /* that this class contains (compile-time) reflected members.  */REFLIB_LF \
    /* Since this operator will never be defined the risk that     */REFLIB_LF \
    /* runtime code uses this conversion erroneously is quite      */REFLIB_LF \
    /* small :-)                                                   */REFLIB_LF \
    operator CompileTIterableObjectTag()const;                       REFLIB_LF \
public:     /* (public) enables the usage of all compile-time */     REFLIB_LF \
            /* variables for all compiletimefunctions here - */      REFLIB_LF \
            /* assume that this does not breach encapsulation */     REFLIB_LF \
            /* because the average programmer would not want to */   REFLIB_LF \
            /* mess with these details anyways */                    REFLIB_LF \
    struct Gap_Num{};                                                REFLIB_LF \
    struct ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END;     REFLIB_LF \
    struct ERROR_ERROR_ERROR_not_a_valid_tag_ERROR_END;              REFLIB_LF \
    struct ERROR_ERROR_ERROR_not_a_valid_baseclass_num_ERROR_END;    REFLIB_LF \
    template<int No, int Dummy = 0> struct GetNameTag                REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END REFLIB_LF \
        type;                                                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int No, int Dummy = 0> struct GetValueType              REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END REFLIB_LF \
        type;                                                        REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END REFLIB_LF \
        ref_type;                                                    REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END REFLIB_LF \
        const_ref_type;                                              REFLIB_LF \
    };                                                               REFLIB_LF \
    template<class Tag, int Dummy = 0> struct GetValueTypeByTag      REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_ERROR_END          REFLIB_LF \
        type;                                                        REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_ERROR_END          REFLIB_LF \
        ref_type;                                                    REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_ERROR_END          REFLIB_LF \
        const_ref_type;                                              REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int No, int Dummy = 0> struct GetMemberKind             REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_position_ERROR_END REFLIB_LF \
        type;                                                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<class Tag, int Dummy = 0> struct GetMemberKindByTag     REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ERROR_ERROR_ERROR_not_a_valid_tag_ERROR_END          REFLIB_LF \
        type;                                                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<class Tag, int Dummy = 0> struct AttributePos;          REFLIB_LF \
    /* the general template is choosen for all lines that do not   */REFLIB_LF \
    /* contain an attribute-defining macro                         */REFLIB_LF \
    template<int Line, int Dummy = 0> struct AttributeNum            REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef Gap_Num IsGap;                                       REFLIB_LF \
        typedef AttributeNum<Line - 1, Dummy> PrevOrGap;             REFLIB_LF \
        enum {PrevIsGap = ::boost::is_same<typename PrevOrGap::IsGap,REFLIB_LF \
                                        Gap_Num>::value};            REFLIB_LF \
        /* find the last attributenum that was defined */            REFLIB_LF \
        typedef typename boost::ct_if<(PrevIsGap !=0),               REFLIB_LF \
                                    typename PrevOrGap::PrevNoGap,   REFLIB_LF \
                                    PrevOrGap>::type PrevNoGap;      REFLIB_LF \
        enum { value = PrevNoGap::value + 1};                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct AttributeNum<0, Dummy>                REFLIB_LF \
    {                                                                REFLIB_LF \
        enum { value = -1};                                          REFLIB_LF \
        typedef int IsGap;                                           REFLIB_LF \
        typedef AttributeNum<0, Dummy> PrevNoGap;                    REFLIB_LF \
    };                                                               REFLIB_LF \
                                                                     REFLIB_LF \
    template<int No, int Dummy = 0> struct GetBaseClass              REFLIB_LF \
    {                                                                REFLIB_LF \
       typedef ERROR_ERROR_ERROR_not_a_valid_baseclass_num_ERROR_END REFLIB_LF \
        type;                                                        REFLIB_LF \
        enum {isVirtual = false};                                    REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Line, int Dummy = 0> struct BaseClassNum            REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef Gap_Num IsGap;                                       REFLIB_LF \
        typedef BaseClassNum<Line - 1, Dummy> PrevOrGap;             REFLIB_LF \
        enum {PrevIsGap = ::boost::is_same<typename PrevOrGap::IsGap,REFLIB_LF \
                                        Gap_Num>::value};            REFLIB_LF \
        /* find the last attributenum that was defined */            REFLIB_LF \
        typedef typename boost::ct_if<(PrevIsGap !=0),               REFLIB_LF \
                                    typename PrevOrGap::PrevNoGap,   REFLIB_LF \
                                    PrevOrGap>::type PrevNoGap;      REFLIB_LF \
        enum { value = PrevNoGap::value + 1};                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct BaseClassNum<0, Dummy>                REFLIB_LF \
    {                                                                REFLIB_LF \
        enum { value = -1};                                          REFLIB_LF \
        typedef int IsGap;                                           REFLIB_LF \
        typedef BaseClassNum<0, Dummy> PrevNoGap;                    REFLIB_LF \
    };                                                               REFLIB_LF \
                                                                     REFLIB_LF \
    typedef ClassName ThisClass;                                     REFLIB_LF \
    enum {beginLineNo = __LINE__};

#define BEGIN_REFLECTION(ClassName)                                  REFLIB_LF \
BEGIN_REFLECTION_NO_CTOR(ClassName)                                  REFLIB_LF \
    ClassName::ClassName()                                           REFLIB_LF \
    {                                                                REFLIB_LF \
        initReflectedMembers();                                      REFLIB_LF \
    }                                                                REFLIB_LF \

#define DEF_REFLECTED_ATTRIBUTE_NO_SETTER_IMPL(ValueType,TagName,MemberKind)    REFLIB_LF \
    struct TagName{};                                                REFLIB_LF \
    enum {TagName##GappedNum = __LINE__ - beginLineNo};              REFLIB_LF \
    template<int Dummy> struct AttributeNum<TagName##GappedNum,      REFLIB_LF \
                                            Dummy>                   REFLIB_LF \
    {                                                                REFLIB_LF \
        /* here I only need a type that is different from Gap_Num */ REFLIB_LF \
        typedef int IsGap;                                           REFLIB_LF \
        typedef AttributeNum<TagName##GappedNum-1, Dummy> PrevOrGap; REFLIB_LF \
        enum {PrevIsGap = ::boost::is_same<typename PrevOrGap::IsGap,REFLIB_LF \
                                            Gap_Num>::value};        REFLIB_LF \
        typedef typename boost::ct_if<(PrevIsGap!=0),                REFLIB_LF \
                                    typename PrevOrGap::PrevNoGap,   REFLIB_LF \
                                    PrevOrGap >::type                REFLIB_LF \
        PrevNoGap;                                                   REFLIB_LF \
                                                                     REFLIB_LF \
        enum { value =  PrevNoGap::value + 1};                       REFLIB_LF \
    };                                                               REFLIB_LF \
    enum {TagName##Num = AttributeNum<TagName##GappedNum>::value};   REFLIB_LF \
    template<int Dummy> struct GetNameTag<TagName##Num, Dummy>       REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef TagName type;                                        REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct GetValueType<TagName##Num, Dummy>     REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ValueType type;                                      REFLIB_LF \
        typedef ::boost::add_reference<ValueType >::type ref_type;   REFLIB_LF \
        typedef ::boost::add_const<type>::type const_type;           REFLIB_LF \
        typedef ::boost::add_reference<const_type>::type             REFLIB_LF \
        const_ref_type;                                              REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct GetValueTypeByTag<TagName, Dummy>     REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef ValueType type;                                      REFLIB_LF \
        typedef ::boost::add_reference<ValueType >::type ref_type;   REFLIB_LF \
        typedef ::boost::add_const<type>::type const_type;           REFLIB_LF \
        typedef ::boost::add_reference<const_type>::type             REFLIB_LF \
        const_ref_type;                                              REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct GetMemberKind<TagName##Num, Dummy>    REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef refl::MemberKind type;                               REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct GetMemberKindByTag<TagName, Dummy>    REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef refl::MemberKind type;                               REFLIB_LF \
    };                                                               REFLIB_LF \
    template<int Dummy> struct AttributePos<TagName, Dummy>          REFLIB_LF \
    {                                                                REFLIB_LF \
        enum {value = TagName##Num};                                 REFLIB_LF \
    };                                                               REFLIB_LF \
    /* not as a fully specialized template because this is quite   */REFLIB_LF \
    /* clumsy to do with MSVC++ 6.0 (the trick with default        */REFLIB_LF \
    /* function arguments does not work)                           */REFLIB_LF \
    static const std::string& getAttributeName(TagName)              REFLIB_LF \
    {                                                                REFLIB_LF \
        static const std::string attName(#TagName);                  REFLIB_LF \
        return attName;                                              REFLIB_LF \
    }                                                                REFLIB_LF \
public:                                                              REFLIB_LF \

#define DEF_REFLECTED_ATTRIBUTE_NO_SETTER(ValueType, TagName)        REFLIB_LF \
DEF_REFLECTED_ATTRIBUTE_NO_SETTER_IMPL(ValueType,TagName,            REFLIB_LF \
                                    InstanceVariableTag)             REFLIB_LF \
    ValueType REFLIBMEMVARNAME(TagName);                             REFLIB_LF \
    GetValueTypeByTag<TagName>::const_ref_type REFLIB_GET_FUNCTION_NAME(TagName)() const  REFLIB_LF \
    { return  REFLIBMEMVARNAME(TagName); }                           REFLIB_LF \
    GetValueTypeByTag<TagName>::ref_type REFLIB_GET_FUNCTION_NAME(TagName)() REFLIB_LF \
    { return  REFLIBMEMVARNAME(TagName); }                           REFLIB_LF \
    GetValueTypeByTag<TagName>::const_ref_type                       REFLIB_LF \
    getValueByTag(TagName)const                                      REFLIB_LF \
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
private:                                                             REFLIB_LF \
    GetValueTypeByTag<TagName>::const_ref_type getByTag(TagName)const REFLIB_LF\
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
    GetValueTypeByTag<TagName>::ref_type getByTag(TagName)           REFLIB_LF \
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
public:

#define DEF_REFLECTED_STATIC_ATTRIBUTE_NO_SETTER(ValueType, TagName) REFLIB_LF \
DEF_REFLECTED_ATTRIBUTE_NO_SETTER_IMPL(ValueType,TagName,            REFLIB_LF \
                                    ClassVariableTag)                REFLIB_LF \
private:                                                             REFLIB_LF \
    static ValueType REFLIBMEMVARNAME(TagName);                      REFLIB_LF \
public:                                                              REFLIB_LF \
    static GetValueTypeByTag<TagName>::const_ref_type REFLIB_GET_FUNCTION_NAME(TagName)() REFLIB_LF \
    { return  REFLIBMEMVARNAME(TagName); }                           REFLIB_LF \
    static GetValueTypeByTag<TagName>::ref_type REFLIB_GET_FUNCTION_NAME(TagName##Ref)()  REFLIB_LF \
    { return  REFLIBMEMVARNAME(TagName); }                           REFLIB_LF \
    static GetValueTypeByTag<TagName>::const_ref_type                REFLIB_LF \
    getValueByTag(TagName)                                           REFLIB_LF \
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
private:                                                             REFLIB_LF \
    static GetValueTypeByTag<TagName>::const_ref_type                REFLIB_LF \
    getByTag(TagName)                                                REFLIB_LF \
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
    static GetValueTypeByTag<TagName>::ref_type getRefByTag(TagName) REFLIB_LF \
    {return REFLIBMEMVARNAME(TagName);}                              REFLIB_LF \
public:

#define DEF_REFLECTED_ATTRIBUTE(ValueType, TagName)                  REFLIB_LF \
DEF_REFLECTED_ATTRIBUTE_NO_SETTER(ValueType,TagName)                 REFLIB_LF \
    void REFLIB_SET_FUNCTION_NAME(TagName)                           REFLIB_LF \
            (GetValueTypeByTag<TagName>::const_ref_type newValue)    REFLIB_LF \
    {                                                                REFLIB_LF \
        setValueByTag(newValue, TagName());                          REFLIB_LF \
    }                                                                REFLIB_LF \
protected:                                                           REFLIB_LF \
    void setByTagImpl                                                REFLIB_LF \
            (GetValueTypeByTag<TagName>::const_ref_type newValue,    REFLIB_LF \
            TagName)                                                 REFLIB_LF \
    {                                                                REFLIB_LF \
        REFLIBMEMVARNAME(TagName) = newValue;                        REFLIB_LF \
    }                                                                REFLIB_LF \
public:                                                              REFLIB_LF \

#define DEF_REFLECTED_STATIC_ATTRIBUTE(ValueType,TagName)            REFLIB_LF \
DEF_REFLECTED_STATIC_ATTRIBUTE_NO_SETTER(ValueType,TagName)          REFLIB_LF \
    static void REFLIB_SET_FUNCTION_NAME(TagName)                    REFLIB_LF \
            (GetValueTypeByTag<TagName>::const_ref_type newValue)    REFLIB_LF \
    {                                                                REFLIB_LF \
        setByTag(newValue, TagName(),refl::ClassVariableTag());      REFLIB_LF \
    }                                                                REFLIB_LF \
protected:                                                           REFLIB_LF \
    static void setByTagImpl                                         REFLIB_LF \
            (GetValueTypeByTag<TagName>::const_ref_type newValue,    REFLIB_LF \
            TagName)                                                 REFLIB_LF \
    {                                                                REFLIB_LF \
        REFLIBMEMVARNAME(TagName) = newValue;                        REFLIB_LF \
    }                                                                REFLIB_LF \
    static void                                                      REFLIB_LF \
    setByTag(GetValueTypeByTag<TagName>::const_ref_type val,         REFLIB_LF \
                         TagName, refl::ClassVariableTag)            REFLIB_LF \
    {                                                                REFLIB_LF \
        /* allthough VC does not need the cast it is still required*/REFLIB_LF \
        /* sets of overloaded functions cannot be used for template*/REFLIB_LF \
        /* parameter deduction.                                    */REFLIB_LF \
        typedef GetValueTypeByTag<TagName>::type MemberType;         REFLIB_LF \
        typedef GetValueTypeByTag<TagName>::const_ref_type           REFLIB_LF \
                      SetterArgType;                                 REFLIB_LF \
        typedef void (*ThisStaticSetImpl)(SetterArgType,TagName);    REFLIB_LF \
        refl::SetWithPreAndPost<ThisClass,SetterArgType,             REFLIB_LF \
                                refl::ClassVariableTag,TagName>      REFLIB_LF \
        wrappedSet(static_cast<ThisStaticSetImpl>(setByTagImpl),val);REFLIB_LF \
        wrappedSet();                                                REFLIB_LF \
    }                                                                REFLIB_LF \
public:                                                              REFLIB_LF \

#define DEF_REFLECTED_VIRTUAL_BASECLASS(BaseClassName)               REFLIB_LF \
DEF_REFLECTED_BASECLASS_IMPL(BaseClassName,true)

#define DEF_REFLECTED_BASECLASS(BaseClassName)                       REFLIB_LF \
DEF_REFLECTED_BASECLASS_IMPL(BaseClassName,false)

#define DEF_REFLECTED_BASECLASS_IMPL(BaseClassName,isVirt)           REFLIB_LF \
    enum {BaseClassName##GappedNum = __LINE__ - beginLineNo};        REFLIB_LF \
    template<int Dummy> struct BaseClassNum<BaseClassName##GappedNum,REFLIB_LF \
                                            Dummy>                   REFLIB_LF \
    {                                                                REFLIB_LF \
        /* here I only need a type that is different from Gap_Num */ REFLIB_LF \
        typedef int IsGap;                                           REFLIB_LF \
        typedef BaseClassNum<BaseClassName##GappedNum-1, Dummy>      REFLIB_LF \
                                PrevOrGap;                           REFLIB_LF \
        enum {PrevIsGap = ::boost::is_same<typename PrevOrGap::IsGap,REFLIB_LF \
                                            Gap_Num>::value};        REFLIB_LF \
        typedef typename boost::ct_if<(PrevIsGap !=0),               REFLIB_LF \
                                    typename PrevOrGap::PrevNoGap,   REFLIB_LF \
                                    PrevOrGap >::type                REFLIB_LF \
        PrevNoGap;                                                   REFLIB_LF \
                                                                     REFLIB_LF \
        enum { value =  PrevNoGap::value + 1};                       REFLIB_LF \
    };                                                               REFLIB_LF \
    enum {BaseClassName##Num =                                       REFLIB_LF \
                    BaseClassNum<BaseClassName##GappedNum>::value};  REFLIB_LF \
    template<int Dummy> struct GetBaseClass<BaseClassName##Num,      REFLIB_LF \
                                            Dummy>                   REFLIB_LF \
    {                                                                REFLIB_LF \
        typedef BaseClassName type;                                  REFLIB_LF \
        enum { isVirtual =  isVirt};                                 REFLIB_LF \
    };                                                               REFLIB_LF \

#define END_REFLECTION                                               REFLIB_LF \
    enum {endLineNo = __LINE__};                                     REFLIB_LF \
    enum {NumAttributes =AttributeNum<endLineNo-beginLineNo>::value};REFLIB_LF \
    enum {NumBaseClasses=BaseClassNum<endLineNo-beginLineNo>::value};REFLIB_LF \
private:                                                             REFLIB_LF \
    template<class Tag> typename GetValueTypeByTag<Tag>::ref_type    REFLIB_LF \
    getByTag(Tag tag, refl::InstanceVariableTag)                     REFLIB_LF \
    {                                                                REFLIB_LF \
        return getByTag(tag);                                        REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Tag> static                                       REFLIB_LF \
    typename GetValueTypeByTag<Tag>::ref_type                        REFLIB_LF \
    getRefByTag(Tag tag, refl::ClassVariableTag)                     REFLIB_LF \
    {                                                                REFLIB_LF \
        return getRefByTag(tag);                                     REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Tag> typename GetValueTypeByTag<Tag>::ref_type    REFLIB_LF \
    getByTag()                                                       REFLIB_LF \
    {                                                                REFLIB_LF \
        return getByTag(Tag(), GetMemberKindByTag<Tag>::type());     REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Tag>                                              REFLIB_LF \
    typename GetValueTypeByTag<Tag>::const_ref_type getByTag()const  REFLIB_LF \
    {                                                                REFLIB_LF \
        return getByTag(Tag(), GetMemberKindByTag<Tag>::type());     REFLIB_LF \
    }                                                                REFLIB_LF \
public:                                                              REFLIB_LF \
    template<class Tag>                                              REFLIB_LF \
    typename GetValueTypeByTag<Tag>::const_ref_type                  REFLIB_LF \
    getByTag(Tag tag, refl::InstanceVariableTag)const                REFLIB_LF \
    {                                                                REFLIB_LF \
        return getByTag(tag);                                        REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Tag>                                              REFLIB_LF \
    static typename GetValueTypeByTag<Tag>::ref_type                 REFLIB_LF \
    getByTag(Tag tag, refl::ClassVariableTag)                        REFLIB_LF \
    {                                                                REFLIB_LF \
        return getRefByTag(tag);                                     REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Tag>                                              REFLIB_LF \
    typename GetValueTypeByTag<Tag>::const_ref_type                  REFLIB_LF \
    getValueByTag()const                                             REFLIB_LF \
    {                                                                REFLIB_LF \
        return getValueByTag(Tag());                                 REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Val,class Tag> void                               REFLIB_LF \
    setValueByTag(Val val, Tag tag)                                  REFLIB_LF \
    {                                                                REFLIB_LF \
        setByTag(val, tag, GetMemberKindByTag<Tag>::type());         REFLIB_LF \
    }                                                                REFLIB_LF \
    template<class Val,class Tag> void                               REFLIB_LF \
    setByTag(Val val, Tag, refl::InstanceVariableTag)                REFLIB_LF \
    {   /* allthough VC does not need the cast it is still required*/REFLIB_LF \
        /* sets of overloaded functions cannot be used for template*/REFLIB_LF \
        /* parameter deduction.                                    */REFLIB_LF \
        typedef typename GetValueTypeByTag<Tag>::type MemberType;    REFLIB_LF \
        typedef typename GetValueTypeByTag<Tag>::const_ref_type      REFLIB_LF \
                      SetterArgType;                                 REFLIB_LF \
        typedef void (ThisClass::*ThisSetImpl)(SetterArgType,Tag);   REFLIB_LF \
        ThisSetImpl setImplPtr = &ThisClass::setByTagImpl;           REFLIB_LF \
        refl::SetWithPreAndPost<ThisClass,MemberType,                REFLIB_LF \
                                refl::InstanceVariableTag,Tag>       REFLIB_LF \
        wrappedSet(*this,                                            REFLIB_LF \
            static_cast<ThisSetImpl>(&ThisClass::setByTagImpl),val); REFLIB_LF \
        wrappedSet();                                                REFLIB_LF \
    }


#endif //
