////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_DEPTHFIRSTFOREACH_H_INC_
#define REFLIB_DEPTHFIRSTFOREACH_H_INC_

#include <boost/type_traits/remove_cv.hpp>

#include <boost/type_traits/remove_pointer.hpp>

#include <iterator>

#include "../qualifiedtype.h"

#include "../itertype.h"

#include "../isruntimeiterable.h"

#include "../recordtags.h"

#include "../priviligedaccess.h"

#include "../baseclassiter.h"

#include "../conditional_end.h"

#include "../conditional_lock.h"

struct NotCtNotRtIterableTag
{};

namespace refl
{
    // for binary traversions when an attribute is a collection
    struct PastEndTag{};
    // we use tag dispatching because with this it is more easy to 
    // define a function object that has default behaviour on each
    // "tagged event" of object-graph traversion.
    struct AttributeEventTag
    {};

    struct EnterAttributeTag:AttributeEventTag
    {};

    struct LeaveAttributeTag:AttributeEventTag
    {};

    struct EnterCollectionTag:RunTIterableObjectTag
    {};

    struct LeaveCollectionTag:RunTIterableObjectTag
    {};

    struct EnterReflectedClassTag:CompileTIterableObjectTag
    {};

    struct LeaveReflectedClassTag:CompileTIterableObjectTag
    {};

    struct EnterCollItemTag
    {};

    struct LeaveCollItemTag
    {};
    template<class T> struct GetIterability
    {
        typedef typename boost::ct_if<(IsRuntimeIterable<T>::value != 0),
                    RunTIterableObjectTag,
                    typename boost::ct_if<(IsCompiletimeIterable<T>::value!= 0),
                                CompileTIterableObjectTag,
                                NotCtNotRtIterableTag>::type>
                    ::type
            type;
    };
    struct IncludeBaseClassesOfAttributesTag{};

    struct DontIncludeBaseClassesOfAttributesTag{};

    template<class ReflectedClass,class MemberKinds,
            class IncludeBaseClassesOfAttributes>
    struct DepthFirst
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};

        template<int Pos> struct SingleAttributeHelper:
                        MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                                                SingleAttributeHelper<Pos> >
        {
            typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                SingleAttributeHelper<Pos> > Inherited;
            using Inherited::processSingleAttribute;
            typedef typename Inherited::AttributeTag CurrentAttributeTag;
            typedef typename Inherited::AttributeValueType CurrentValueType;
            // deep traversions follow pointers - 
            // example: 
            // when the type of a member variable is a pointer to an 
            // std::container of pointers to reflected classes then
            // the same graph event call backs are invoked as when the type
            // of the member variable is an std::container of instances of 
            // a reflected class.
            // In particular the leaf functions are called with dereferenced
            // pointers to the reflected class - container items.
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& attribute,
                            Function& function,ReflectedClass& recordData, 
                            AttributeTag attributeTag,TypeIsNoPointerTag)
            {
                processNode(attribute,function,recordData, attributeTag,
                    GetIterability<ValueType>::type());
            }

            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& lhsAttribute,
                            ValueType& rhsAttribute,Function& function,
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            AttributeTag attributeTag,TypeIsNoPointerTag)
            {
                processNode(lhsAttribute, rhsAttribute, function,lhsRecord, rhsRecord, attributeTag,
                            GetIterability<ValueType>::type());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& attribute,
                            Function& function,ReflectedClass& recordData, 
                            AttributeTag attributeTag,TypeIsPointerTag)
            {
                typedef typename boost::remove_pointer<ValueType>::type ValueTypeDeReferenced;
                if(attribute)
                    processDeReferenced(*attribute,function,recordData,attributeTag,
                                IsPointer<ValueTypeDeReferenced>::type());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& lhsAttribute,
                            ValueType& rhsAttribute,Function& function,
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            AttributeTag attributeTag,TypeIsPointerTag)
            {
                typedef typename boost::remove_pointer<ValueType>::type ValueTypeDeReferenced;
                if(lhsAttribute)
                {
                    if(rhsAttribute)
                        processDeReferenced(*lhsAttribute,*rhsAttribute, function,
                                        lhsRecord,rhsRecord,attributeTag,
                                        IsPointer<ValueTypeDeReferenced>::type());
                    else
                        function(*lhsAttribute, PastEndTag(), attributeTag, 
                                &lhsRecord, &rhsRecord);

                }
                else
                {
                    if(rhsAttribute)
                        function(PastEndTag(), *rhsAttribute, attributeTag, 
                                &lhsRecord, &rhsRecord);
                }
            }
            // defines what has to happen, when the current attribute kind 
            // matches MemberKinds
            template<class Function> static void 
            processMatchingAttribute(ReflectedClass& recordData,
                                    Function& function)
            {  
#ifdef _DEBUG   // depending on the abilities of your debugger this may help 
                // to answer the question: "where the hell am I now?"
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                typename Inherited::RefType valueRef=Inherited::getValueHandle(recordData);
                function(valueRef, CurrentAttributeTag(), EnterAttributeTag(), 
                    &recordData);

                processDeReferenced(valueRef,function,recordData,CurrentAttributeTag(),
                                IsPointer<CurrentValueType>::type());

                function(valueRef, CurrentAttributeTag(), LeaveAttributeTag(), 
                    &recordData);
            }
            // binary overload
            template<class Function> static void 
            processMatchingAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                                    Function& function)
            {  
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                typename Inherited::RefType lhsValueRef=Inherited::getValueHandle(lhsRecord);
                typename Inherited::RefType rhsValueRef=Inherited::getValueHandle(rhsRecord);
                function(lhsValueRef, rhsValueRef, CurrentAttributeTag(), 
                    EnterAttributeTag(), &lhsRecord, &rhsRecord);

                processDeReferenced(lhsValueRef,rhsValueRef,function,lhsRecord,rhsRecord,
                            CurrentAttributeTag(),IsPointer<CurrentValueType>::type());

                function(lhsValueRef, rhsValueRef, CurrentAttributeTag(), LeaveAttributeTag(), 
                    &lhsRecord, &rhsRecord);
            }

            template<class CollectionType, class Function, class AttributeTag> 
            static void foreach(CollectionType& attribute, Function& function, 
                            ReflectedClass& recordData,AttributeTag attributeTag,
                            DontCheckEndCondition)
            {
                typedef typename GetIterType<CollectionType>::type Iter;
                typedef typename CollectionType::value_type ValueType;
                Iter pos = attribute.begin();
                Iter end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function(*pos, attributeTag, EnterCollItemTag(), 
                            &recordData);
                    processDeReferenced(*pos,function,recordData,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*pos,attributeTag, LeaveCollItemTag(),  
                            &recordData);
                }
            }
            template<class CollectionType, class Function, class AttributeTag> 
            static void foreach(CollectionType& attribute, Function& function, 
                            ReflectedClass& recordData, AttributeTag attributeTag,
                            CheckEndCondition)
            {
                typedef typename GetIterType<CollectionType>::type Iter;
                typedef typename CollectionType::value_type ValueType;
                Iter pos = attribute.begin();
                Iter end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function(*pos, attributeTag, EnterCollItemTag(), 
                            &recordData);
                    processDeReferenced(*pos,function,recordData,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*pos,attributeTag, LeaveCollItemTag(),  
                            &recordData);
                    if(function.done())
                        return;
                }
            }
            // the attribute tagged with AttributeTag may be a collection of 
            // collections (arbitrarily nested). This function gets called if 
            // the current type is a collection. Especially if 
            // CollectionType::value_type is a collection as well, then this 
            // function template (of course with other template arguments) will 
            // recursively get called again from the inner loop.
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& attribute, Function& function, 
                                ReflectedClass& recordData,AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                // At least with Visual C++ 7.1 and below I can't have these
                // typedefs at class - scope.
                typedef typename boost::remove_cv<Function>::type Undecorated;
                typedef typename HasConditionalEnd<bool Undecorated::*>::type ConditionalBreak;

                function(attribute, attributeTag, EnterCollectionTag(),
                        &recordData);
                foreach(attribute,function,recordData,attributeTag, ConditionalBreak());
                function(attribute, attributeTag, LeaveCollectionTag(),
                        &recordData);
            }   
            // binary overload - if an attribute is a collection then it 
            // may happen that the sizes of the Collections differ
            // function() may therefor have to handle
            // the case when lhsItem is undefined or rhsItem is undefined
            template<class Iter, class Function, class AttributeTag> 
            static void foreach(Iter& lhsPos, const Iter& lhsEnd,
                            Iter& rhsPos, const Iter& rhsEnd,
                            Function& function, 
                            ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                            AttributeTag attributeTag,
                            DontCheckEndCondition)
            {
                typedef typename std::iterator_traits<Iter>::value_type ValueType;
                for(; lhsPos != lhsEnd && rhsPos != rhsEnd ; ++lhsPos, ++rhsPos)
                {
                    function(*lhsPos, *rhsPos, attributeTag, EnterCollItemTag(), 
                            &lhsRecord, &rhsRecord);
                    processDeReferenced(*lhsPos, *rhsPos,function,
                                lhsRecord, rhsRecord,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*lhsPos, *rhsPos,attributeTag, LeaveCollItemTag(),  
                            &lhsRecord, &rhsRecord);
                }
            }
            template<class Iter, class Function, class AttributeTag> 
            static void foreach(Iter& lhsPos, const Iter& lhsEnd,
                            Iter& rhsPos, const Iter& rhsEnd,
                            Function& function, 
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                            AttributeTag attributeTag,
                            CheckEndCondition)
            {
                typedef typename std::iterator_traits<Iter>::value_type ValueType;
                for(; lhsPos != lhsEnd && rhsPos != rhsEnd ; ++lhsPos, ++rhsPos)
                {
                    function(*lhsPos, *rhsPos, attributeTag, EnterCollItemTag(), 
                            &lhsRecord, &rhsRecord);
                    processDeReferenced(*lhsPos, *rhsPos,function,
                                lhsRecord, rhsRecord,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*lhsPos, *rhsPos,attributeTag, LeaveCollItemTag(),  
                            &lhsRecord, &rhsRecord);
                    if(function.done())
                        return;
                }
            }
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& lhsAttribute,CollectionType& rhsAttribute, 
                                Function& function, 
                                ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                // At least with Visual C++ 7.1 and below I can't have these
                // typedefs at class - scope.
                typedef typename boost::remove_cv<Function>::type Undecorated;
                typedef typename HasConditionalEnd<bool Undecorated::*>::type ConditionalBreak;
                typedef typename GetIterType<CollectionType>::type Iter;
                function(lhsAttribute,rhsAttribute, attributeTag, EnterCollectionTag(),
                        &lhsRecord, &rhsRecord);
                Iter lhsPos = lhsAttribute.begin();
                Iter lhsEnd = lhsAttribute.end();
                Iter rhsPos = rhsAttribute.begin();
                Iter rhsEnd = rhsAttribute.end();
                foreach(lhsPos,lhsEnd,rhsPos,rhsEnd,function,
                    lhsRecord,rhsRecord,attributeTag, ConditionalBreak());
                if(lhsPos == lhsEnd)
                {
                    // this decision may be wrong - binary traversions stop when
                    // corresponding items are not available
                    // (Currently the only use case I have for a binary traversion
                    // is generating comparison operators - for that this decision
                    // fits the bill).
                    // Apart from that - the function called below is free 
                    // to start a new unary traversion with whatever attribute function
                    if(rhsPos != rhsEnd)
                        function(PastEndTag(), rhsPos, rhsEnd, attributeTag, 
                                &lhsRecord, &rhsRecord);
                }
                else
                    function(lhsPos, lhsEnd, PastEndTag(), attributeTag, 
                            &lhsRecord, &rhsRecord);
                function(lhsAttribute,rhsAttribute, attributeTag, LeaveCollectionTag(),
                        &lhsRecord, &rhsRecord);
            }   

            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, 
                                    Function& function, 
                                    ReflectedClass& recordData,
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function(attribute, attributeTag, EnterReflectedClassTag(), 
                        &recordData);
                depthFirstForeach(attribute,function,MemberKinds(),
                                IncludeBaseClassesOfAttributes());
                function(attribute, attributeTag, LeaveReflectedClassTag(), 
                        &recordData);
            }
            // binary overload
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& lhsAttribute, ValueType& rhsAttribute, 
                                    Function& function, 
                                    ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function(lhsAttribute, rhsAttribute, attributeTag, EnterReflectedClassTag(), 
                        &lhsRecord, &rhsRecord);
                depthFirstForeach(lhsAttribute, rhsAttribute,function,MemberKinds(),
                                IncludeBaseClassesOfAttributes());
                function(lhsAttribute, rhsAttribute, attributeTag, LeaveReflectedClassTag(), 
                        &lhsRecord, &rhsRecord);
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, Function& function, 
                                ReflectedClass& recordData, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {  
                function(attribute, attributeTag, &recordData);   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }
            // binary overload - do something with corresponding attributes of lhsRecord
            // and rhsRecord
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& lhsAttribute, ValueType& rhsAttribute, 
                                Function& function, 
                                ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {  
                function(lhsAttribute, rhsAttribute, attributeTag, &lhsRecord, &rhsRecord);   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }
        };
        template<class Function, int RestToDo> struct Iter: 
        SingleAttributeHelper<TotalNumAttributes - RestToDo>
        {
            static void goon(ReflectedClass& recordData, Function& function, CheckEndCondition)
            {
                if(!function.done())
                    Iter<Function,RestToDo-1>::process(recordData,function);
            }
            static void goon(ReflectedClass& recordData, Function& function, DontCheckEndCondition)
            {
                Iter<Function,RestToDo-1>::process(recordData,function);
            }
            static void process(ReflectedClass& recordData,Function& function)
            {
                // At least with Visual C++ 7.1 and below I can't have these
                // typedefs at class - scope.
                typedef typename boost::remove_cv<Function>::type Undecorated;
                typedef typename HasConditionalEnd<bool Undecorated::*>::type ConditionalBreak;
                Inherited::processSingleAttribute(recordData,function);
                goon(recordData,function,ConditionalBreak());
            }
            // binary overloads.
            static void goon(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function, CheckEndCondition)
            {
                if(!function.done())
                    Iter<Function,RestToDo-1>::process(lhsRecord, rhsRecord,function);
            }
            static void goon(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function, DontCheckEndCondition)
            {
                Iter<Function,RestToDo-1>::process(lhsRecord, rhsRecord,function);
            }
            static void process(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,Function& function)
            {
                typedef typename boost::remove_cv<Function>::type Undecorated;
                typedef typename HasConditionalEnd<bool Undecorated::*>::type ConditionalBreak;
                Inherited::processSingleAttribute(lhsRecord, rhsRecord,function);
                goon(lhsRecord, rhsRecord,function,ConditionalBreak());
            }
        };
        template<class Function> struct Iter<Function,1>: 
        SingleAttributeHelper<TotalNumAttributes - 1>
        {
            static void process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
            }
            static void process(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,Function& function)
            {
                Inherited::processSingleAttribute(lhsRecord, rhsRecord,function);
            }
        };
        // having support for baseclasses changes in the reflected members could
        // otherwise "silently" change the compilability of derived classes.
        template<class Function> struct Iter<Function,0>
        {
            static void process(ReflectedClass&, Function&)
            {}
        };
        template<class Function> static void
        doForeach(ReflectedClass& dataTuple, Function& function)
        {
            ConditionalLock<ReflectedClass> lock(dataTuple); 
            Iter<Function,TotalNumAttributes>::process(dataTuple,function);
        }
        // binary traversion (currently only supported when instance members
        // shall be considered as well).
        template<class Function> static void
        doForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function)
        {
            ConditionalLock<ReflectedClass> lhsLock(lhsRecord); 
            ConditionalLock<ReflectedClass> rhsLock(rhsRecord); 
            Iter<Function,TotalNumAttributes>::process(lhsRecord,rhsRecord,function);
        }
    };

    // to traverse static members only having no instance of ReflectedClass
    template<class ReflectedClass, class IncludeBaseClassesOfAttributes>
    struct DepthFirst<ReflectedClass,ClassVariableTag,
                    IncludeBaseClassesOfAttributes>
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};

        template<int Pos> struct SingleAttributeHelper:
                        MatchingAttributeHandler<ReflectedClass,
                            ClassVariableTag,Pos,SingleAttributeHelper<Pos> >
        {
            typedef MatchingAttributeHandler<ReflectedClass,ClassVariableTag,
                Pos,SingleAttributeHelper<Pos> > Inherited;
            using Inherited::processSingleAttribute;
            typedef typename Inherited::AttributeTag CurrentAttributeTag;
            typedef typename Inherited::AttributeValueType CurrentValueType;
            // defines what has to happen, when the current attribute kind 
            // matches ClassVariableTag
            template<class Function> static void 
            processMatchingAttribute(Function& function)
            {  
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                typename Inherited::RefType valueRef=Inherited::getValueHandle();
                function.template operator()<ReflectedClass>(valueRef, 
                                    CurrentAttributeTag(), EnterAttributeTag());

                processNode(valueRef,function,CurrentAttributeTag(),
                            GetIterability<CurrentValueType>::type());

                function.template operator()<ReflectedClass>(valueRef, 
                                    CurrentAttributeTag(), LeaveAttributeTag());
            }

            // the attribute tagged with AttributeTag may be a collection of 
            // collections (arbitrarily nested). This function gets called if 
            // the current type is a collection. Especially if 
            // CollectionType::value_type is a collection as well, then this 
            // function template (of course with other template arguments) will 
            // recursively get called again from the inner loop.
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& attribute, 
                                Function& function, 
                                AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                function.template operator()<ReflectedClass>(attribute, 
                            attributeTag, EnterCollectionTag());
                typedef typename GetIterType<CollectionType>::type Iter;
                Iter pos = attribute.begin();
                Iter end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function.template operator()<ReflectedClass>(*pos, 
                                            attributeTag, EnterCollItemTag());
                    processNode(*pos,function,attributeTag,
                            GetIterability<BOOST_DEDUCED_TYPENAME
                                        CollectionType::value_type>::type());
                    function.template operator()<ReflectedClass>(*pos, 
                                            attributeTag, LeaveCollItemTag());
                }
                function.template operator()<ReflectedClass>(attribute,  
                                            attributeTag, LeaveCollectionTag());
            }   

            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, 
                                    Function& function, 
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function.template operator()<ReflectedClass>(attribute,  
                                    attributeTag, EnterReflectedClassTag());
                // currently I cannot think of an example where only the static
                // state of a static member variable should be considered here
                // if so - an additional template parameter 
                // (like MemberMemberKinds) might need to be added to the 
                // outermost enclosing class template of this function.
                depthFirstForeach(attribute,function,
                                TraverseInstanceAndClassVariables(),
                                IncludeBaseClassesOfAttributes());
                function.template operator()<ReflectedClass>(attribute,  
                                        attributeTag, LeaveReflectedClassTag());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, Function& function, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {
                function.template operator()<ReflectedClass>(attribute, 
                                                            attributeTag);   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }
        };
        template<class Function, int RestToDo> struct Iter: 
        SingleAttributeHelper<TotalNumAttributes - RestToDo>
        {
            static void process(Function& function)
            {
                Inherited::processSingleAttribute(function);
                Iter<Function,RestToDo-1>::process(function);
            }
        };
        template<class Function> struct Iter<Function,1>: 
        SingleAttributeHelper<TotalNumAttributes - 1>
        {
            static void process(Function& function)
            {
                Inherited::processSingleAttribute(function);
            }
        };
        // having support for baseclasses changes in the reflected members could
        // otherwise "silently" change the compilability of derived classes.
        template<class Function> struct Iter<Function,0>
        {
            static void process(Function&)
            {}
        };
        template<class Function> static void
        doForeach(Function& function)
        {
            ConditionalLock<ReflectedClass> lock; 
            Iter<Function,TotalNumAttributes>::process(function);
        }
    };
    template<bool IncludeBaseClasses> struct IncludeBases;

    template<> struct IncludeBases<false>
    {
        typedef DontIncludeBaseClassesOfAttributesTag MemberTraversionType;
    };

    template<> struct IncludeBases<true>
    {
        typedef IncludeBaseClassesOfAttributesTag MemberTraversionType;
    };
    // the ~PassTag functions might need to be moved to a namespace
    // detail.
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& dataTuple, Function& function,
                            MemberKinds,DontIncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,
                                                            function);
    }
    // binary (ReflectedClass) overload
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, Function& function,
                            MemberKinds,DontIncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(lhsRecord, rhsRecord,
                                                            function);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& dataTuple, Function& function,
                                MemberKinds,IncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            IncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,function);
    }

    // binary (ReflectedClass) overload
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, Function& function,
                                MemberKinds,IncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            IncludeBaseClassesOfAttributesTag>::doForeach(lhsRecord, rhsRecord,function);
    }
    // having this as a function object facilitates its use as parameter
    // to other compile-time recursions.
    template<class Function, class MemberKinds, 
            bool IncludeBaseClassesOfMembers> struct 
    DeepAttributeFunction :IncludeBases<IncludeBaseClassesOfMembers>
    {
        typedef IncludeBases<IncludeBaseClassesOfMembers> Inherited;
        DeepAttributeFunction(Function& function):function_(function)
        {}
        template<class ReflectedClass> void
        operator()(ReflectedClass& dataTuple)const
        {
            function_(dataTuple);    // useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag(dataTuple,function_,MemberKinds(),
                                traversionType);
        }
        template<class ReflectedClass> void
        operator()(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord)const
        {
            function_(lhsRecord, rhsRecord);  // may or may not be useful for binary traversions
                                            // this gets called when a base class is entered    
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag(lhsRecord, rhsRecord,function_,MemberKinds(),
                                traversionType);
        }
        Function& function_;
    };
    template<class Function, 
            bool IncludeBaseClassesOfMembers> struct 
    DeepAttributeFunction<Function,ClassVariableTag,IncludeBaseClassesOfMembers>: 
                                    IncludeBases<IncludeBaseClassesOfMembers>
    {
        typedef IncludeBases<IncludeBaseClassesOfMembers> Inherited;
        DeepAttributeFunction(Function& function):function_(function)
        {}
        template<class ReflectedClass> void operator()()const
        {
            function_.template operator()<ReflectedClass>();    
            // useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag<ReflectedClass>(function_,
                                ClassVariableTag(),traversionType);
        }
        // this one will be entered if a static member variable contains
        // reflected attributes - then we do have an instance and still
        // want to traverse classvariables only.
        template<class ReflectedClass> 
        void operator()(ReflectedClass& dataTuple)const
        {
            function_(dataTuple);    // useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag(dataTuple,function_,
                                ClassVariableTag(),traversionType);
        }
        Function& function_;
    };
    // some algorithms may explicitly need to traverse attributes of the
    // baseclasses others don't (for instance if InitReflectedMembers
    // is called in each ctor of each reflected class then this function
    // has to use foreachAttribute as its back-end (to avoid unnecessary
    // assignments (and not foreachAttributeInclBaseClasses)).

    // Note that ReflectedClass may contain a const qualifier (among others)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple,Function& function,
                         MemberKinds)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,
                                                            function);
    }
    // when it is a compiletimevariable whether baseclasses should be 
    // traversed - implementation detail that might help others too.
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple, Function& function,
                    MemberKinds kinds, DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach(dataTuple,function,kinds);
    }
    // binary traversion helper (like for comparisons of ReflectedClasses)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function,
                    MemberKinds kinds, DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach(lhsRecord, rhsRecord,function,kinds);
    }
    // when it is a compiletimevariable whether baseclasses should be 
    // traversed - implementation detail that might help others too.
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple, Function& function, 
                        MemberKinds kinds, IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses(dataTuple,function,kinds);
    }
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,  
                        Function& function, MemberKinds kinds, 
                        IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses(lhsRecord, rhsRecord,function,kinds);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachIncAllBaseClasses(ReflectedClass& dataTuple, 
                                            Function& function,
                                            MemberKinds)
    {
        DeepAttributeFunction<Function,MemberKinds,true> 
                                depthFirstIteration(function);
        foreachBaseClass(dataTuple,depthFirstIteration);
        depthFirstIteration(dataTuple);
    }
    // binary traversion (like for comparisons of ReflectedClasses)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachIncAllBaseClasses(ReflectedClass& lhsRecord, 
                                            ReflectedClass& rhsRecord, 
                                            Function& function,
                                            MemberKinds)
    {
        DeepAttributeFunction<Function,MemberKinds,true> 
                                depthFirstIteration(function);
        foreachBaseClass(lhsRecord, rhsRecord,depthFirstIteration);
        depthFirstIteration(lhsRecord, rhsRecord);
    }
    // for traversing static members when no appropriate instance of 
    // ReflectedClass is available
    // Note that ReflectedClass may contain a const qualifier (among others)
    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function)
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function, ClassVariableTag kinds, 
                            DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach<ReflectedClass>(function,kinds);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function, ClassVariableTag kinds, 
                        IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses<ReflectedClass>(function,kinds);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachPassTag(Function& function, ClassVariableTag kinds, 
                         DontIncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
                    DontIncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachPassTag(Function& function,ClassVariableTag, 
                             IncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
            IncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachIncAllBaseClasses(Function& function)
    {
        DeepAttributeFunction<Function,ClassVariableTag ,true> 
                                depthFirstIteration(function);
        foreachBaseClass<ReflectedClass>(depthFirstIteration);
        depthFirstIteration.template operator()<ReflectedClass>();
    }
}
#endif // !

