////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_FLATFOREACH_H_INC_
#define REFLIB_FLATFOREACH_H_INC_


#include <typeinfo>

#include "../conditional_lock.h"
#include "../baseclassiter.h"
#include "../priviligedaccess.h"

namespace refl
{
    template<class ReflectedClass, class MemberKinds> struct AllAttributes
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};
        template<int Pos> struct SingleAttributeHelper:
                        MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                                                SingleAttributeHelper<Pos> >
        {
            typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                                        SingleAttributeHelper<Pos> > Inherited;
            typedef typename Inherited::AttributeTag CurrentAttributeTag;
            // defines what has to happen, when the current attribute kind
            // matches MemberKinds
            template<class Function> static void 
            processMatchingAttribute(ReflectedClass& recordData,
                                    Function& function)
            {  
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                function(Inherited::getValueHandle(recordData), CurrentAttributeTag(), 
                        &recordData);   
                // This helps for nested attributes - to pass the 
            }   // type of the enclosing class back to the caller
        };
        template<class Function, int RestToDo> struct Iter: 
                           SingleAttributeHelper<TotalNumAttributes - RestToDo> 
        {
            typedef SingleAttributeHelper<TotalNumAttributes - RestToDo>  Inherited;
            static void process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
                Iter<Function,RestToDo-1>::process(recordData,function);
            }
        };
        template<class Function> struct Iter<Function,1>: 
                                SingleAttributeHelper<TotalNumAttributes - 1>
        {
            typedef SingleAttributeHelper<TotalNumAttributes - 1>  Inherited;
            static void process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
            }
        };
        // having support for baseclasses changes in the reflected members could
        // otherwise "silently" change the compilability of derived classes.
        template<class Function> struct Iter<Function,0>
        {
            static void process(ReflectedClass&, Function&)
            {}
        };
        template<class Function> static void
        doForeach(ReflectedClass& dataTuple, Function& function)
        {
            ConditionalLock<ReflectedClass> lock(dataTuple);
            Iter<Function,TotalNumAttributes>::process(dataTuple,function);
        }
    };

    template<class ReflectedClass> struct AllAttributes<ReflectedClass,
                                                            ClassVariableTag>
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};
        template<int Pos> struct SingleAttributeHelper:
                        MatchingAttributeHandler<ReflectedClass,
                            ClassVariableTag,Pos,SingleAttributeHelper<Pos> >
        {
            typedef MatchingAttributeHandler<ReflectedClass,ClassVariableTag,
                                    Pos,SingleAttributeHelper<Pos> > Inherited;
            typedef typename Inherited::AttributeTag CurrentAttributeTag;
            // defines what has to happen, when the current attribute kind
            // matches MemberKinds
            template<class Function> static void
            processMatchingAttribute(Function& function)
            {
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
#endif
                function.template operator()<ReflectedClass>(Inherited::getValueHandle(), CurrentAttributeTag());
                // This helps for nested attributes - to pass the
            }   // type of the enclosing class back to the caller
        };
        template<class Function, int RestToDo> struct Iter: 
                           SingleAttributeHelper<TotalNumAttributes - RestToDo> 
        {
            typedef SingleAttributeHelper<TotalNumAttributes - RestToDo>  Inherited;
            static void process(Function& function)
            {
                Inherited::processSingleAttribute(function);
                Iter<Function,RestToDo-1>::process(function);
            }
        };
        template<class Function> struct Iter<Function,1>: 
                                SingleAttributeHelper<TotalNumAttributes - 1>
        {
            typedef SingleAttributeHelper<TotalNumAttributes - 1>  Inherited;
            static void process(Function& function)
            {
                Inherited::processSingleAttribute(function);
            }
        };
        // having support for baseclasses changes in the reflected members could
        // otherwise "silently" change the compilability of derived classes.
        template<class Function> struct Iter<Function,0>
        {
            static void process(Function&)
            {}
        };
        template<class Function> static void 
        doForeach(Function& function)
        {
            ConditionalLock<ReflectedClass> lock;
            Iter<Function,TotalNumAttributes>::process(function);
        }
    };

    
    // All usages so far can take advantage of the AttributeTag supplied - hence
    // this Tag will be passed as Parameter to the function back
    template<class ReflectedClass, class Function, class MemberKinds> 
    void foreachAttribute(ReflectedClass& dataTuple, 
                        Function& function,
                        MemberKinds)   
    {
        AllAttributes<ReflectedClass,MemberKinds>::doForeach(dataTuple,
                                                            function);
    }
    // without a ReflectedClass instance only static members can be iterated
    template<class ReflectedClass, class Function> 
    void foreachAttribute(Function& function)   
    {
        AllAttributes<ReflectedClass,ClassVariableTag>::doForeach(function);
    }

    template<class Function,class MemberKinds> struct ForeachAttribute;
    // some algorithm may explicitly need to traverse attributes of the 
    // baseclasses others don't (for instance if initReflectedMembers
    // is called in each ctor of each reflected class then this function
    // does has to use foreachAttribute as its back-end (to avoid unnecessary
    // assignments).
    template<class ReflectedClass, class Function, class MemberKinds> 
    void foreachAttributeInclBaseClasses(ReflectedClass& dataTuple, 
                                        Function& function,
                                        MemberKinds)   
    {
        ForeachAttribute<Function,MemberKinds> directMembersIteration(function);
        foreachBaseClass(dataTuple,directMembersIteration);
        directMembersIteration(dataTuple);
    }
    template<class ReflectedClass, class Function> 
    void foreachAttributeInclBaseClasses(Function& function)   
    {
        ForeachAttribute<Function,ClassVariableTag> 
                                            directMembersIteration(function);
        foreachBaseClass<ReflectedClass>(directMembersIteration);
        directMembersIteration.template operator()<ReflectedClass>();
    }
    // having this as a function object facilitates its use as parameter
    // to other compile-time recursions.
    template<class Function, class MemberKinds> struct ForeachAttribute
    {
        ForeachAttribute(Function& function):function_(function)
        {}
        template<class ReflectedClass> void 
        operator()(ReflectedClass& dataTuple)const
        {
            foreachAttribute(dataTuple,function_,MemberKinds());
        }
        Function& function_;
    };

    template<class Function> struct ForeachAttribute<Function,ClassVariableTag>
    {
        ForeachAttribute(Function& function):function_(function)
        {}
        template<class ReflectedClass> void operator()()const
        {
            foreachAttribute<ReflectedClass>(function_);
        }
        Function& function_;
    };
}

#endif
