////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_DEPTHFIRSTFOREACH_H_INC_
#define REFLIB_DEPTHFIRSTFOREACH_H_INC_

#include <iterator>

#include "../itertype.h"

#include "../isruntimeiterable.h"

#include "../recordtags.h"

#include "../priviligedaccess.h"

#include "../baseclassiter.h"

#include "../conditional_end.h"

#include "../conditional_lock.h"

struct NotCtNotRtIterableTag
{};

namespace refl
{
    // for binary traversions when an attribute is a collection or a pointer
    struct PastEndTag{};

    // we use tag dispatching because with this it is more easy to 
    // define a function object that has default behaviour on each
    // "tagged event" of object-graph traversion.
    struct AttributeEventTag
    {};

    struct EnterAttributeTag:AttributeEventTag
    {};

    struct LeaveAttributeTag:AttributeEventTag
    {};

    struct EnterCollectionTag:RunTIterableObjectTag
    {};

    struct LeaveCollectionTag:RunTIterableObjectTag
    {};

    struct EnterReflectedClassTag:CompileTIterableObjectTag
    {};

    struct LeaveReflectedClassTag:CompileTIterableObjectTag
    {};

    struct EnterCollItemTag
    {};

    struct LeaveCollItemTag
    {};


    template<class T> struct GetIterability
    {
        typedef typename boost::ct_if<(IsRuntimeIterable<T>::value != 0),
                    RunTIterableObjectTag,
                    typename boost::ct_if<(IsCompiletimeIterable<T>::value!= 0),
                                CompileTIterableObjectTag,
                                NotCtNotRtIterableTag>::type>
                    ::type
            type;
    };
    struct IncludeBaseClassesOfAttributesTag{};

    struct DontIncludeBaseClassesOfAttributesTag{};

    template<class ReflectedClass,class MemberKinds, 
    class IncludeBaseClassesOfAttributes=DontIncludeBaseClassesOfAttributesTag>
    struct DepthFirst
    {
        enum {TotalNumAttributes = ReflectedClass::NumAttributes};
        template<int Pos> struct SingleAttributeHelper
        {
            typedef SingleAttributeHelper<Pos> ThisClass;
            template<class Function> static void 
            processSingleAttribute(ReflectedClass& recordData,
                                Function& function)
            {  
                typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                    ThisClass> 
                    Handler;
                Handler::processSingleAttribute(recordData,function);
            }
            // binary overload
            template<class Function> static void 
            processSingleAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                                Function& function)
            {  
                typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                    ThisClass> 
                    Handler;
                Handler::processSingleAttribute(lhsRecord,rhsRecord,function);
            }
            // deep traversions follow pointers - 
            // example: 
            // when the type of a member variable is a pointer to an 
            // std::container of pointers to reflected classes then
            // the same graph event call backs are invoked as when the type
            // of the member variable is an std::container of instances of 
            // a reflected class.
            // In particular the leaf functions are called with dereferenced
            // pointers to the reflected class - container items.
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& attribute,
                            Function& function,ReflectedClass& recordData, 
                            AttributeTag attributeTag,TypeIsNoPointerTag)
            {
                processNode(attribute,function,recordData, attributeTag,
                    GetIterability<ValueType>::type());
            }

            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& lhsAttribute,
                            ValueType& rhsAttribute,Function& function,
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            AttributeTag attributeTag,TypeIsNoPointerTag)
            {
                processNode(lhsAttribute, rhsAttribute, function,lhsRecord, rhsRecord, attributeTag,
                            GetIterability<ValueType>::type());
            }
            // overload because boost::remove_pointer requires partial spec of class templates
            template<class ValueTypeDeReferenced, class Function, class AttributeTag> 
            static void processDeReferenced(ValueTypeDeReferenced& attribute,
                            Function& function,ReflectedClass& recordData, 
                            AttributeTag attributeTag)
            {
                processDeReferenced(attribute,function,recordData,attributeTag,
                            IsPointer<ValueTypeDeReferenced>::type());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& attribute,
                            Function& function,ReflectedClass& recordData, 
                            AttributeTag attributeTag,TypeIsPointerTag)
            {
                if(attribute)
                    processDeReferenced(*attribute,function,recordData,attributeTag);
            }
            // overload (binary case) because boost::remove_pointer requires partial spec of class templates
            template<class ValueTypeDeReferenced, class Function, class AttributeTag> 
            static void processDeReferenced(ValueTypeDeReferenced& lhsAttribute,
                            ValueTypeDeReferenced& rhsAttribute,Function& function,
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            AttributeTag attributeTag)
            {
                processDeReferenced(lhsAttribute,rhsAttribute, function,lhsRecord,rhsRecord,attributeTag,
                                IsPointer<ValueTypeDeReferenced>::type());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processDeReferenced(ValueType& lhsAttribute,
                            ValueType& rhsAttribute,Function& function,
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                            AttributeTag attributeTag,TypeIsPointerTag)
            {
                if(lhsAttribute)
                {
                    if(rhsAttribute)
                        processDeReferenced(*lhsAttribute,*rhsAttribute, 
                                function,lhsRecord,rhsRecord,attributeTag);
                    else
                        function(*lhsAttribute, PastEndTag(), attributeTag, 
                                PassScope<ReflectedClass>(&lhsRecord), 
                                PassScope<ReflectedClass>(&rhsRecord));

                }
                else
                {
                    if(rhsAttribute)
                        function(PastEndTag(), *rhsAttribute, attributeTag, 
                            PassScope<ReflectedClass>(&lhsRecord), 
                            PassScope<ReflectedClass>(&rhsRecord));
                }
            }
            template<class Function, class ValueType> static void 
            processMatchingAttribute(ReflectedClass& recordData,
                                Function& function, ValueType& attribute)
            {  
                // MSVC++ 6.0 does not allow this typedef at class scope.
                // Hence accessing a value of a reflected class at a given 
                // position is implemented per copy and paste of this line
                typedef typename ReflectedClass::template GetNameTag<Pos,
                    ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
                const char* ValueTypeName = typeid(ValueType).name();
#endif
                function(attribute, CurrentAttributeTag(), EnterAttributeTag(),
                        PassScope<ReflectedClass>(&recordData));

                processDeReferenced(attribute,function,recordData,CurrentAttributeTag(),
                                IsPointer<ValueType>::type());

                function(attribute, CurrentAttributeTag(), LeaveAttributeTag(), 
                        PassScope<ReflectedClass>(&recordData));
            }
            // binary overload
            template<class Function, class ValueType> static void 
            processMatchingAttribute(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                                Function& function, ValueType& lhsAttribute, ValueType& rhsAttribute)
            {  
                // MSVC++ 6.0 does not allow this typedef at class scope.
                // Hence accessing a value of a reflected class at a given 
                // position is implemented per copy and paste of this line
                typedef typename ReflectedClass::template GetNameTag<Pos,
                    ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
                const char* ValueTypeName = typeid(ValueType).name();
#endif
                function(lhsAttribute, rhsAttribute, CurrentAttributeTag(), EnterAttributeTag(),
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));

                processDeReferenced(lhsAttribute, rhsAttribute,function,lhsRecord,rhsRecord,
                            CurrentAttributeTag(),IsPointer<ValueType>::type());

                function(lhsAttribute, rhsAttribute, CurrentAttributeTag(), LeaveAttributeTag(), 
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
            }
              
            template<class CollectionType, class Function, class AttributeTag> 
            static void foreach(CollectionType& attribute, Function& function, 
                            ReflectedClass& recordData,AttributeTag attributeTag,
                            DontCheckEndCondition)
            {
                typedef typename GetIterType<CollectionType>::type IterType;
                typedef typename CollectionType::value_type ValueType;
                IterType pos = attribute.begin();
                IterType end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function(*pos, attributeTag, EnterCollItemTag(), 
                            PassScope<ReflectedClass>(&recordData));
                    processDeReferenced(*pos,function,recordData,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*pos,attributeTag, LeaveCollItemTag(),  
                            PassScope<ReflectedClass>(&recordData));
                }
            }
            template<class CollectionType, class Function, class AttributeTag> 
            static void foreach(CollectionType& attribute, Function& function, 
                            ReflectedClass& recordData, AttributeTag attributeTag,
                            CheckEndCondition)
            {
                typedef typename GetIterType<CollectionType>::type IterType;
                typedef typename CollectionType::value_type ValueType;
                IterType pos = attribute.begin();
                IterType end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function(*pos, attributeTag, EnterCollItemTag(), 
                            PassScope<ReflectedClass>(&recordData));
                    processDeReferenced(*pos,function,recordData,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*pos,attributeTag, LeaveCollItemTag(),  
                            PassScope<ReflectedClass>(&recordData));
                    if(function.done())
                        return;
                }
            }

            // the attribute tagged with AttributeTag may be a collection of 
            // collections (arbitrarily nested). This function gets called if 
            // the current type is a collection. Especially if 
            // CollectionType::value_type is a collection as well, then this 
            // function template (of course with other template arguments) will 
            // recursively get called again from the inner loop.
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& attribute, 
                                Function& function, 
                                ReflectedClass& recordData,
                                AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                typedef typename HasConditionalEnd<Function>::type ConditionalBreak;
                function(attribute, attributeTag, EnterCollectionTag(),
                        PassScope<ReflectedClass>(&recordData));
                foreach(attribute,function,recordData,attributeTag, ConditionalBreak());
                function(attribute,  attributeTag, LeaveCollectionTag(),
                        PassScope<ReflectedClass>(&recordData));
            }   

            // binary overload - if an attribute is a collection then it 
            // may happen that the sizes of the Collections differ
            // function() may therefor have to handle
            // the case when lhsItem is undefined or rhsItem is undefined
            template<class IterType, class Function, class AttributeTag, class ValueType> 
            static void foreach(IterType& lhsPos, const IterType& lhsEnd,
                            IterType& rhsPos, const IterType& rhsEnd,
                            Function& function, 
                            ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                            AttributeTag attributeTag,
                            DontCheckEndCondition,
                            PassType<ValueType>)    // workaround for incomplete iterator_traits
            {
                for(; lhsPos != lhsEnd && rhsPos != rhsEnd ; ++lhsPos, ++rhsPos)
                {
                    function(*lhsPos, *rhsPos, attributeTag, EnterCollItemTag(), 
                            PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                    processDeReferenced(*lhsPos, *rhsPos,function,
                                lhsRecord, rhsRecord,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*lhsPos, *rhsPos,attributeTag, LeaveCollItemTag(),  
                            PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                }
            }
            template<class IterType, class Function, class AttributeTag, class ValueType> 
            static void foreach(IterType& lhsPos, const IterType& lhsEnd,
                            IterType& rhsPos, const IterType& rhsEnd,
                            Function& function, 
                            ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                            AttributeTag attributeTag,
                            CheckEndCondition,
                            PassType<ValueType>)    // workaround for incomplete iterator_traits
            {
                for(; lhsPos != lhsEnd && rhsPos != rhsEnd ; ++lhsPos, ++rhsPos)
                {
                    function(*lhsPos, *rhsPos, attributeTag, EnterCollItemTag(), 
                            PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                    processDeReferenced(*lhsPos, *rhsPos,function,
                                lhsRecord, rhsRecord,attributeTag,
                                IsPointer<ValueType>::type());
                    function(*lhsPos, *rhsPos,attributeTag, LeaveCollItemTag(),  
                            PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                    if(function.done())
                        return;
                }
            }
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& lhsAttribute,CollectionType& rhsAttribute, 
                                Function& function, 
                                ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                typedef typename HasConditionalEnd<Function>::type ConditionalBreak;
                typedef typename GetIterType<CollectionType>::type IterType;
                typedef typename GetQualifiedMemberType<CollectionType,
                                            BOOST_DEDUCED_TYPENAME CollectionType::value_type>::type 
                                            ValueType;
                function(lhsAttribute,rhsAttribute, attributeTag, EnterCollectionTag(),
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                IterType lhsPos = lhsAttribute.begin();
                IterType lhsEnd = lhsAttribute.end();
                IterType rhsPos = rhsAttribute.begin();
                IterType rhsEnd = rhsAttribute.end();
                foreach(lhsPos,lhsEnd,rhsPos,rhsEnd,function,
                    lhsRecord,rhsRecord,attributeTag, ConditionalBreak(),PassType<ValueType>());
                if(lhsPos == lhsEnd)
                {
                    // this decision may be wrong - binary traversions stop when
                    // corresponding items are not available
                    // (Currently the only use case I have for a binary traversion
                    // is generating comparison operators - for that this decision
                    // fits the bill).
                    // Apart from that - the function called below is free 
                    // to start a new unary traversion with whatever attribute function
                    if(rhsPos != rhsEnd)
                        function(PastEndTag(), rhsPos, rhsEnd, attributeTag, 
                                PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                }
                else
                    function(lhsPos, lhsEnd, PastEndTag(), attributeTag, 
                            PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                function(lhsAttribute,rhsAttribute, attributeTag, LeaveCollectionTag(),
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
            }   

            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, 
                                    Function& function, 
                                    ReflectedClass& recordData,
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function(attribute,  attributeTag, EnterReflectedClassTag(), 
                        PassScope<ReflectedClass>(&recordData));
                depthFirstForeach(attribute,function,
                                MemberKinds(),IncludeBaseClassesOfAttributes());
                function(attribute,  attributeTag, LeaveReflectedClassTag(), 
                        PassScope<ReflectedClass>(&recordData));
            }
            // binary overload
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& lhsAttribute, ValueType& rhsAttribute, 
                                    Function& function, 
                                    ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function(lhsAttribute, rhsAttribute, attributeTag, EnterReflectedClassTag(), 
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
                depthFirstForeach(lhsAttribute, rhsAttribute,function,MemberKinds(),
                                IncludeBaseClassesOfAttributes());
                function(lhsAttribute, rhsAttribute, attributeTag, LeaveReflectedClassTag(), 
                        PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));
            }

            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, Function& function, 
                                ReflectedClass& recordData, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {  
                function(attribute, attributeTag, PassScope<ReflectedClass>(&recordData));   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }
            // binary overload - do something with corresponding attributes of lhsRecord
            // and rhsRecord
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& lhsAttribute, ValueType& rhsAttribute, 
                                Function& function, 
                                ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {  
                function(lhsAttribute, rhsAttribute, attributeTag, PassScope<ReflectedClass>(&lhsRecord), PassScope<ReflectedClass>(&rhsRecord));   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }

            // overlaods for access to static members without having an
            // instance of reflected class.
            template<class Function> static void 
            processSingleAttribute(Function& function)
            {  
                typedef MatchingAttributeHandler<ReflectedClass,MemberKinds,Pos,
                    ThisClass> 
                    Handler;
                Handler::processSingleAttribute(function);
            }
            template<class Function, class ValueType> static void 
            processMatchingAttribute(Function& function, ValueType& attribute)
            {  
                typedef typename ReflectedClass::template GetNameTag<Pos,
                    ReflectedClass::ThisClass>::type 
                    CurrentAttributeTag;
#ifdef _DEBUG
                const char* AttributeName = typeid(CurrentAttributeTag).name();
                const char* ValueTypeName = typeid(ValueType).name();
#endif
                function(attribute, CurrentAttributeTag(), EnterAttributeTag(), 
                        PassType<ReflectedClass>());

                processNode(attribute,function,CurrentAttributeTag(),
                            GetIterability<ValueType>::type());

                function(attribute, CurrentAttributeTag(), LeaveAttributeTag(), 
                        PassType<ReflectedClass>());
            }

            // the attribute tagged with AttributeTag may be a collection of 
            // collections (arbitrarily nested). This function gets called if 
            // the current type is a collection. Especially if 
            // CollectionType::value_type is a collection as well, then this 
            // function template (of course with other template arguments) will 
            // recursively get called again from the inner loop.
            template<class CollectionType, class Function, class AttributeTag> 
            static void processNode(CollectionType& attribute, 
                                Function& function, 
                                AttributeTag attributeTag,
                                RunTIterableObjectTag runTimeCollTag)
            {
                function(attribute, attributeTag, EnterCollectionTag(),PassType<ReflectedClass>());
                typedef typename GetIterType<CollectionType>::type IterType;
                IterType pos = attribute.begin();
                IterType end = attribute.end();
                for(; pos != end; ++pos)
                {
                    function(*pos, attributeTag, EnterCollItemTag(), 
                            PassType<ReflectedClass>());
                    processNode(*pos,function,recordData,attributeTag,
                            GetIterability<BOOST_DEDUCED_TYPENAME 
                                        CollectionType::value_type>::type());
                    function(*pos, attributeTag, LeaveCollItemTag(), 
                            PassType<ReflectedClass>());
                }
                function(attribute, attributeTag, LeaveCollectionTag(),PassType<ReflectedClass>());
            }   

            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, 
                                    Function& function, 
                                    AttributeTag attributeTag, 
                                    CompileTIterableObjectTag)
            {  
                function(attribute,  attributeTag, EnterReflectedClassTag(), 
                        PassType<ReflectedClass>());
                // currently I cannot think of an example where only the static
                // state of a static member variable should be considered here
                // if so - an additional template parameter 
                // (like MemberMemberKinds) might need to be added to the 
                // outermost enclosing class template of this function.
                depthFirstForeach(attribute,function,
                                TraverseInstanceAndClassVariables(),
                                IncludeBaseClassesOfAttributes());
                function(attribute,  attributeTag, LeaveReflectedClassTag(), 
                        PassType<ReflectedClass>());
            }
            template<class ValueType, class Function, class AttributeTag> 
            static void processNode(ValueType& attribute, Function& function, 
                                AttributeTag attributeTag, 
                                NotCtNotRtIterableTag)
            {  
                function(attribute, attributeTag, PassType<ReflectedClass>());   
                // This helps for nested attributes - especially to pass the 
                // type of the enclosing class back to the caller.
            }
        };
        template<int RestToDo> struct Iter:
                          SingleAttributeHelper<TotalNumAttributes - RestToDo> 
        {
            template<class Function> static void 
                goon(ReflectedClass& recordData, Function& function, CheckEndCondition)
            {
                if(!function.done())
                    Iter<RestToDo-1>::process(recordData,function);
            }
            template<class Function> static void 
                goon(ReflectedClass& recordData, Function& function, DontCheckEndCondition)
            {
                Iter<RestToDo-1>::process(recordData,function);
            }
            typedef SingleAttributeHelper<TotalNumAttributes - RestToDo> Inherited;
            template<class Function> static void 
            process(ReflectedClass& recordData,Function& function)
            {
                typedef typename HasConditionalEnd<Function>::type ConditionalBreak;
                Inherited::processSingleAttribute(recordData,function);
                goon(recordData,function,ConditionalBreak());
            }
            // binary overloads.
            template<class Function> static void 
                goon(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function, CheckEndCondition)
            {
                if(!function.done())
                    Iter<RestToDo-1>::process(lhsRecord, rhsRecord,function);
            }
            template<class Function> static void 
                goon(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function, DontCheckEndCondition)
            {
                Iter<RestToDo-1>::process(lhsRecord, rhsRecord,function);
            }
            template<class Function> static void 
            process(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,Function& function)
            {
                typedef typename HasConditionalEnd<Function>::type ConditionalBreak;
                Inherited::processSingleAttribute(lhsRecord, rhsRecord, function);
                goon(lhsRecord, rhsRecord,function,ConditionalBreak());
            }
            template<class Function> static void 
            process(Function& function)
            {
                Inherited::processSingleAttribute(function);
                Iter<RestToDo-1>::process(function);
            }
        };
        // this is illegal in standard c++ : full specialization definition
        // inside the body of a primary template 
        template<> struct Iter<1>: 
        SingleAttributeHelper<TotalNumAttributes - 1>
        {
            typedef SingleAttributeHelper<TotalNumAttributes - 1> Inherited;
            template<class Function> static void 
            process(ReflectedClass& recordData,Function& function)
            {
                Inherited::processSingleAttribute(recordData,function);
            }
            template<class Function> static void 
            process(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,Function& function)
            {
                Inherited::processSingleAttribute(lhsRecord, rhsRecord,function);
            }
            template<class Function> static void 
            process(Function& function)
            {
                Inherited::processSingleAttribute(function);
            }
        };
        template<> struct Iter<0>
        {
            template<class Function> static void 
                process(ReflectedClass& ,Function& )
            {}
            template<class Function> static void 
                process(ReflectedClass&,ReflectedClass&,Function& )
            {}
            template<class Function> static void process(Function& )
            {}
        };
        template<class Function> static void 
        doForeach(ReflectedClass& dataTuple, Function& function)
        {
            ConditionalLock<ReflectedClass> lock(dataTuple); 
            Iter<TotalNumAttributes>::process(dataTuple,function);
        }
        template<class Function> static void 
        doForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function)
        {
            ConditionalLock<ReflectedClass> lhsLock(lhsRecord); 
            ConditionalLock<ReflectedClass> rhsLock(rhsRecord); 
            Iter<TotalNumAttributes>::process(lhsRecord, rhsRecord,function);
        }
        template<class Function> static void 
        doForeach(Function& function)
        {
            ConditionalLock<ReflectedClass> lock; 
            Iter<TotalNumAttributes>::process(function);
        }
    };

    template<bool IncludeBaseClasses> struct IncludeBases;

    template<> struct IncludeBases<false>
    {
        typedef DontIncludeBaseClassesOfAttributesTag MemberTraversionType;
    };

    template<> struct IncludeBases<true>
    {
        typedef IncludeBaseClassesOfAttributesTag MemberTraversionType;
    };

    // the ~PassTag functions might need to be moved to a namespace
    // detail.
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& dataTuple, Function& function,
                         MemberKinds,DontIncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,
                                                            function);
    }

    // binary (ReflectedClass) overload
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, Function& function,
                         MemberKinds,DontIncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(lhsRecord, rhsRecord,
                                                            function);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& dataTuple, Function& function,
                             MemberKinds,IncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            IncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,function);
    }

    // binary (ReflectedClass) overload
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachPassTag(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, Function& function,
                             MemberKinds,IncludeBaseClassesOfAttributesTag)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            IncludeBaseClassesOfAttributesTag>::doForeach(lhsRecord, rhsRecord,function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachPassTag(Function& function, ClassVariableTag kinds, 
                         DontIncludeBaseClassesOfAttributesTag, 
                         PassType<ReflectedClass> = PassType<ReflectedClass>())
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
                DontIncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachPassTag(Function& function,ClassVariableTag, 
                        IncludeBaseClassesOfAttributesTag, 
                        PassType<ReflectedClass> = PassType<ReflectedClass>())
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
            IncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    // having this as a function object facilitates its use as parameter
    // to other compile-time recursions.
    template<class Function, class MemberKinds,  
        bool IncludeBaseClassesOfMembers> struct 
    DeepAttributeFunction :IncludeBases<IncludeBaseClassesOfMembers>
    {
        typedef IncludeBases<IncludeBaseClassesOfMembers> Inherited;
        DeepAttributeFunction(Function& function):function_(function)
        {}
        // this can be entered if a static member variable contains
        // reflected attributes and MemberKinds is ClassVariableTag
        // - then we do have an instance and still want to traverse 
        // classvariables only.
        template<class ReflectedClass> void
        operator()(ReflectedClass& dataTuple)const
        {
            function_(dataTuple);    // useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag(dataTuple,function_,MemberKinds(),
                                    traversionType);
        }
        // binary overload
        template<class ReflectedClass> void
        operator()(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord)const
        {
            function_(lhsRecord, rhsRecord);    // was useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag(lhsRecord, rhsRecord,function_,MemberKinds(),
                                    traversionType);
        }
        // this function should reside in a partial specialization
        // of DeepAttributeFunction specialized on MemberKinds = ClassVariableTag
        template<class ReflectedClass> void
        operator()(PassType<ReflectedClass> type, 
                ClassVariableTag classVars)const
        {
            function_(type,classVars);    // useful when printing a baseclass
            typename Inherited::MemberTraversionType traversionType;
            depthFirstForeachPassTag<ReflectedClass>(function_,classVars,
                                                    traversionType);
        }
        Function& function_;
    };
    // some algorithms may explicitly need to traverse attributes of the
    // baseclasses others don't (for instance if InitReflectedMembers
    // is called in each ctor of each reflected class then this function
    // has to use foreachAttribute as its back-end (to avoid unnecessary
    // assignments (and not foreachAttributeInclBaseClasses)).

    // Note that ReflectedClass may contain a const qualifier (among others)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple,Function& function,
                         MemberKinds)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(dataTuple,
                                                            function);
    }
    // binary overload
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,
                        Function& function, MemberKinds)
    {
        DepthFirst<ReflectedClass,MemberKinds,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(lhsRecord, rhsRecord,
                                                            function);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple, Function& function,
                        MemberKinds kinds, 
                        DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach(dataTuple,function,kinds);
    }

    // binary traversion helper (like for comparisons of ReflectedClasses)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, Function& function,
                    MemberKinds kinds, DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach(lhsRecord, rhsRecord,function,kinds);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& dataTuple, Function& function,
                        MemberKinds kinds, IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses(dataTuple,function,kinds);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord,  
                        Function& function, MemberKinds kinds, 
                        IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses(lhsRecord, rhsRecord,function,kinds);
    }

    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachIncAllBaseClasses(ReflectedClass& dataTuple, 
                                           Function& function,
                                           MemberKinds)
    {
        DeepAttributeFunction<Function,MemberKinds,true> 
                                depthFirstIteration(function);
        foreachBaseClass(dataTuple,depthFirstIteration);
        depthFirstIteration(dataTuple);
    }
    // binary traversion (like for comparisons of ReflectedClasses)
    template<class ReflectedClass, class Function, class MemberKinds>
    void depthFirstForeachIncAllBaseClasses(ReflectedClass& lhsRecord, 
                                            ReflectedClass& rhsRecord, 
                                            Function& function,
                                            MemberKinds)
    {
        DeepAttributeFunction<Function,MemberKinds,true> 
                                depthFirstIteration(function);
        foreachBaseClass(lhsRecord, rhsRecord,depthFirstIteration);
        depthFirstIteration(lhsRecord, rhsRecord);
    }

    // for traversing static members when no appropriate instance of 
    // ReflectedClass is available
    // Note that ReflectedClass may contain a const qualifier (among others)
    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function)
    {
        DepthFirst<ReflectedClass,ClassVariableTag,
            DontIncludeBaseClassesOfAttributesTag>::doForeach(function);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function, ClassVariableTag kinds, 
                            DontIncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeach<ReflectedClass>(function,kinds);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeach(Function& function, ClassVariableTag kinds, 
                        IncludeBaseClassesOfAttributesTag)
    {
        depthFirstForeachIncAllBaseClasses<ReflectedClass>(function,kinds);
    }

    template<class ReflectedClass, class Function>
    void depthFirstForeachIncAllBaseClasses(Function& function)
    {
        DeepAttributeFunction<Function,ClassVariableTag ,true> 
                                depthFirstIteration(function);
        foreachBaseClass<ReflectedClass>(depthFirstIteration);
        depthFirstIteration(PassType<ReflectedClass>(),ClassVariableTag());
    }
}
#endif // !

