////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_USERCONFIG_H_INC_
#define REFLIB_USERCONFIG_H_INC_
#include <boost/utility.hpp>
#include <boost/pending/ct_if.hpp>
#include "boost/type_traits/is_same.hpp"

// use this form if you prefer member variables with trailing_ underscore
#define REFLIB_MEMVARS_WITH_TRAILING_
// use this form if you prefer member variables with leading m_
// #define REFLIB_MEMVARS_WITH_LEADING_m_
// you cannot use both forms

// use this macro to customize the name of the setfunctions that contain
// the tagname as substring.
// Note that the templated set functions are not configurable because
// these may be used by generative algorithms (depthfirstforeach, ...).
// Catering for variable function names in generative algorithms 
// would add complexity where already is more than enough of that.
#define REFLIB_SET_FUNCTION_NAME(TagName) set##TagName

#define REFLIB_GET_FUNCTION_NAME(TagName) get##TagName
// the stl-style would be: (relying on overloading)
// #define REFLIB_SET_FUNCTION_NAME(TagName) TagName
// #define REFLIB_GET_FUNCTION_NAME(TagName) TagName


namespace refl
{
    struct SingleThreadingModelTag{};

    struct MultiThreadingModelTag{};

    struct EmptyMutex:boost::noncopyable {};

    // each lock shall be constructible with its corresponding mutex
    struct EmptyLock:boost::noncopyable 
    {
        EmptyLock(EmptyMutex&){}
    };
}

// having both: base class and attribute traversion inside one function 
// may generate an internal compiler error with VC++ 6.0
// With this define the set of traversed virtual base classes will be
// calculated at runtime.
#define REFLIB_USE_RUNTIME_VBASE_FILTER

// if REFLIB_USE_RUNTIME_VBASE_FILTER is defined then you can choose whether
// you want to have the set of visited virtual generated once (and stored
// in a set of function static variables) or recalculated on each traversion
// of baseclasses of each given class. (If you want to have the latter
// behaviour uncomment the following define).
#define REFLIB_STORE_TRAVERSED_VBASES_STATIC

namespace userconfig
{
    // This can be useful for instance to apply additional checks on the new 
    // value 
    template<class ReflectedClass, class AttributeTag, class ValueType> 
    void preSet(ReflectedClass& rec, AttributeTag tag, ValueType val)
    { /* whatever needs to be done */}

    // the same for static member variables
    template<class ReflectedClass, class AttributeTag, class ValueType> 
    void preSet(AttributeTag tag, ValueType val)
    { /* whatever needs to be done */}

    // This was useful because the records did store the count of a container in
    // a separate variable setting the count-member therefore had to resize the
    // associated collection. (So this function can be used to ensure that post
    // conditions hold.)
    template<class ReflectedClass, class AttributeTag> void
    postSet(ReflectedClass& rec, AttributeTag)
    { /* whatever needs to be done */}

    // the same for static member variables
    template<class ReflectedClass, class AttributeTag> void
    postSet(AttributeTag)
    { /* whatever needs to be done */}

    
    typedef refl::SingleThreadingModelTag ThreadingModel;
    //typedef refl::MultiThreadingModelTag ThreadingModel;

    // the settings for single threading might not need user adjustment
    template<class ReflectedClass> struct SingleThreadingMutexType
    {
        typedef refl::EmptyMutex type;
    };

    template<class ReflectedClass> struct SingleThreadingLockType
    {
        typedef refl::EmptyLock type;
    };

    // the settings for multi threading DO NEED USER ADJUSTMENT
    template<class ReflectedClass> struct MultiThreadingMutexType
    {
        typedef refl::EmptyMutex type;
    };

    template<class ReflectedClass> struct MultiThreadingLockType
    {
        typedef refl::EmptyLock type;
    };

    template<class ReflectedClass, class ThreadingModelType> struct MutexType
    {
        enum {isSingleThreading = boost::is_same<ThreadingModelType,
                                        refl::SingleThreadingModelTag>::value};
        typedef typename boost::ct_if<(isSingleThreading == 0),
                            typename SingleThreadingMutexType<ReflectedClass>::type,
                            typename MultiThreadingMutexType<ReflectedClass>::type>::type
            type;
    };

    template<class ReflectedClass, class ThreadingModelType> struct LockType
    {
        enum {isSingleThreading = boost::is_same<ThreadingModelType,
                                        refl::SingleThreadingModelTag>::value};
        typedef typename boost::ct_if<(isSingleThreading == 0),
                            typename SingleThreadingLockType<ReflectedClass>::type,
                            typename MultiThreadingLockType<ReflectedClass>::type>::type
            type;
    };

    // This one can be used to have one mutex per class (these mutexes will be 
    // used for setting static member variables.
    template<class ReflectedClass, class ThreadingModelType> class MutexInstance
    {
    public:
        typedef typename MutexType<ReflectedClass,ThreadingModelType>::type 
            Mutex;
        static Mutex& get();
    private:
        static Mutex mutex_;
    };
    template <class ReflectedClass, class ThreadingModelType> 
    typename MutexInstance<ReflectedClass,ThreadingModelType>::Mutex&
    MutexInstance<ReflectedClass,ThreadingModelType>::get()
    {
        return mutex_;
    }

    template <class ReflectedClass, class ThreadingModelType>
    typename MutexInstance<ReflectedClass,ThreadingModelType>::Mutex 
    MutexInstance<ReflectedClass,ThreadingModelType>::mutex_;
}

#endif //
